/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.

  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#ifndef _TWA2835_H_
#define _TWA2835_H_ 1

/*! \brief This structure holds the status for a video input channel */
struct twa2835_status {
  /*! detected format:
    0 PAL-B/D
    1 PAL-M
    2 PAL-N
    3 PAL-60
    4 NTSC-M
    5 NTSC-4.43
    6 NTSC-N
  */
  int det_format;
  int det_color;		/*!< color detected  */
  int lock_color;		/*!< locked color */
  int lock_gain;		/*!< locked autogain  */
  int lock_ofst;		/*!< locked offser */
  int lock_hpll;		/*!< locked hpll */
  int det_nonstd;		/*!< non-zero if input source is non standard */
  int det_fld60;		/*!< non-zero if input source is 60 Hz */
  int id;			/*!< TWA2835 chip id */
  int ver;			/*!< TWA2835 chip revision */
};

/*! \brief Event status for each for each channel from one A/B mux  */
struct twa2835_motion_substatus {
  char night[4];		/*!< entry i is non-zero if night vision detected on channel i */
  char blind[4];		/*!< entry i is non-zero if blind detected on channel i  */
  char motion[4];		/*!< entry i is non-zero if motion detected on channel i  */
  char loss[4];			/*!< entry i is non-zero if loss of signal on channel i */
};

/*! \brief Event status for all the TWA2835 inputs */
struct twa2835_motion_status {
  struct twa2835_motion_substatus vina;	/*!< event status for mux A  */
  struct twa2835_motion_substatus vinb; /*!< event status for mux B  */
};

/*! \brief Pixel to be displayed on the monitor/record path oof the TWA2835 */
struct twa2835_pixel {
  int index:6;			/*!< index into the LUT */
  int mix:1;			/*!< mix the pixel with video  */
  int blink:1;			/*!< blink the pixel */
  int off:1;			/*!< pixel is off (transparent)  */
};

/* these are platform specific */

/*! \brief reads the value of a TWA2835 register

  \param twa pointer to the TWA2835 instance

  \param addr address of register to read

  \return value read from register
*/
int twa2835_read_reg(void *twa, int addr);

/*! \brief writes to a TWA2835 register

  \param twa pointer to the TWA2835 instance

  \param addr address of register to write

  \param data value to write
*/
void twa2835_write_reg(void *twa, int addr, int data);

/*! \brief executes a batch of operations to the TWA2835

  \param twa pointer to the TWA2835 instance

  \param batch batch to execute
 */
int twa2835_batch(void *twa, int *batch);

/*! \brief resets the TWA2835

  \param twa pointer to the TWA2835 instance
*/
void twa2835_reset(void *twa);

/*! \brief returns the number of second elapsed from an arbitrary moment in time

  \return the number of second elapsed from an arbitrary moment in time
*/
unsigned long twa2835_get_sec(void);

/*! \brief waits for an event generated by the TWA2835

  \param twa pointer to the TWA2835 instance

  \param sec number of seconds to wait before returning even if no events

  \param nv returns the status of no-video event. Bits 0-3 are
  channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \param md returns the status of motion detection event. Bits 0-3 are
  channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \param bd returns the status of blind detection event. Bits 0-3 are
  channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \param nd returns the status of night detect event. Bits 0-3 are
  channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \return 0 on success, otherwise non-zero
*/
int twa2835_wait_motion(void *twa, int sec, int *nv, int *md, int *bd, int *nd);

/* these are generic */

/*! \brief initialize the hardware associated with the given instance

  \param twa pointer to the TWA2835 instance

  \param isNtsc non-zero to set NTSC mode

  \return 0 on success, otherwise non-zero
*/

int twa2835_init(void *twa, int isNtsc);

/*! \brief releases a TWA2835 instance

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_release(void *twa);

/*! \brief setup video output mode

  \param twa pointer to the TWA2835 instance

  \param tv_x_rec enables if non-zero record path (instead of monitor) on analog X output

  \param tv_x_ovl enables graphic overlay on analog X output

  \param tv_y_rec enables if non-zero record path (instead of monitor) on analog Y output

  \param tv_y_ovl enables graphic overlay on analog Y output

  \param enc_rec enables if non-zero record path (instead of monitor) on digital (ADV212 connected) output

  \param enc_ovl enables graphic overlay on digital (ADV212 connected) output

  \return 0 on success, otherwise non-zero
*/
int twa2835_output_mode(void *twa, int tv_x_rec, int tv_x_ovl, int tv_y_rec, int tv_y_ovl, int enc_rec, int enc_ovl);

/*! \brief set the RGB valuess for an entry in the palette

  \param twa pointer to the TWA2835 instance

  \param ind palette index

  \param R red value (0-255)

  \param G green value (0-255)

  \param B blue value (0-255)

  \return 0 on success, otherwise non-zero
*/
int twa2835_set_palette(void *twa, int ind, int R, int G, int B);

/*! \brief display a channel full-screen on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_full(void *twa, int ch);

/*! \brief display all the 4 channels in the monitor path

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_quad(void *twa);

/*! \brief freezes a channel on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_freeze_ch(void *twa, int ch);

/*! \brief puts a channel in continuos mode on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_continue_ch(void *twa, int ch);

/*! \brief updates a channel when freezed on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_strobe_ch(void *twa, int ch);

/*! \brief puts channel in switched mode on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_switch_mode_ch(void *twa, int ch);

/*! \brief changes a channel in switched mode to the given channel on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_switch_trigger(void *twa, int ch);

/*! \brief freezes the monitor path

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_freeze(void *twa);

/*! \brief puts the monitor path in continuos mode

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_continue(void *twa);

/*! \brief enables/disables playback on a channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param ena non-zero to enable

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_pb_enable(void *twa, int ch, int ena);

/*! \brief shows where a motion has been detected by coloring user boxes

  \param twa pointer to the TWA2835 instance

  \param box user box to use

  \param enable non-zero to enable

  \param ab input to use from AB shitch (0=A, 1=B)

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_show_motion_box(void *twa, int box, int enable, int ab, int ch);

/*! \brief displays a bitmap on the monitor path

  \param twa pointer to the TWA2835 instance

  \param x x coordinate (0-712) multiple of 8

  \param y y coordinate (0-232/280) multiple of 8

  \param w image width (8-720) multiple of 8

  \param h image height (0-240/288) multiple of 8

  \param pn which video page to write (0-5)

  \param p bitmap bits (w*h pixels)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_write_bitmap(void *twa, int x, int y, int w, int h, int pn, struct twa2835_pixel *p);

/*! \brief enable display of bitmap on monitor path

  \param twa pointer to the TWA2835 instance

  \param ena if non-zero do enable

  \param pn video page to use (0-5)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_bitmap_enable(void *twa, int ena, int pn);

/*! \brief enables gathering playback information from ebedded channel id

  \param twa pointer to the TWA2835 instance

  \param auto_vb_detect_ena enable channel id auto detection

  \param vbi_ena enable analog chanel id

  \param vbi_code_ena enable digital channel id

  \param vbi_ric_ena enable run-time clock from channel id

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_pb_id(void *twa, int auto_vb_detect_ena, int vbi_ena, int vbi_code_ena,
		      int vbi_ric_ena);

/*! \brief crops playback area so embedded channel id information are not visible

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param ena if non-zero do enable

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_pb_crop(void *twa, int ch, int ena);

/*! \brief sets the attribute for a channel on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param blank blanks channel if non-zero

  \param hmirror horizontal mirroring of channel if non-zero

  \param vmirror vertical mirroring of channel if non-zero

  \param blink blinks channel boundary channel if non-zero

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_ch_attr(void *twa, int ch, int blank, int hmirror,
			int vmirror, int blink);

/*! \brief enables blinking of channel boundary if video-loss

  \param twa pointer to the TWA2835 instance

  \param ena enable function if non-zero

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_blink_on_loss(void *twa, int ena);

/*! \brief enables 2x zoom on the monitor path

  \param twa pointer to the TWA2835 instance

  \param ena enable function if non-zero

  \param x x coordinate of the top-left corner of zoom region

  \param y y coordinate of the top-left corner of zoom region

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_zoom(void *twa, int ena, int x, int y);

/*! \brief sets the parameter of the anti-nois filter on the monitor path

  \param twa pointer to the TWA2835 instance

  \param mode adaptive (0) or median (1) mode

  \param slope gradient of filter (0-3)

  \param thr threshold of filter (0-31)

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_filter(void *twa, int mode, int slope, int thr);

/*! \brief shows the cursor on the monitor path

  \param twa pointer to the TWA2835 instance

  \param large non-zero for large cursor

  \param filled non-zero for filled cursor

  \param blink non-zero for blinking cursor

  \param x x coordintate, if -1 turns off the cursor

  \param y y coordintate, if -1 turns off the cursor

  \return 0 on success, otherwise non-zero
*/
int twa2835_mon_cursor(void *twa, int large, int filled, int blink, int x, int y);

/*! \brief a channel full-screen on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_full(void *twa, int ch);

/*! \brief display all the 4 channels in the record path

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_quad(void *twa);

/*! \brief displays a bitmap on the record path

  \param twa pointer to the TWA2835 instance

  \param x x coordinate (0-712) multiple of 8

  \param y y coordinate (0-232/280) multiple of 8

  \param w image width (8-720) multiple of 8

  \param h image height (0-240/288) multiple of 8

  \param p bitmap bits (w*h pixels)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_write_bitmap(void *twa, int x, int y, int w, int h, struct twa2835_pixel *p);

/*! \brief enable display of bitmap on the record path

  \param twa pointer to the TWA2835 instance

  \param ena if non-zero do enable

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_bitmap_enable(void *twa, int ena);

/*! \brief enable the outputing of channel id information on the record path

  \param twa pointer to the TWA2835 instance

  \param vis_ena analog channel id

  \param vis_auto_ena automatic channel id

  \param vis_auto_rpt_ena channel id repetition

  \param vis_det_ena detection channel id

  \param vis_user_ena user channel id

  \param vis_code_ena digital channel id

  \param vis_ric_ena  run-time clock

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_id_enable(void *twa,
			  int vis_ena, int vis_auto_ena, int vis_auto_rpt_ena, int vis_det_ena,
			  int vis_user_ena, int vis_code_ena, int vis_ric_ena);

/*! \brief sets the attribute for a channel on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param blank blanks channel if non-zero

  \param hmirror horizontal mirroring of channel if non-zero

  \param vmirror vertical mirroring of channel if non-zero

  \param blink blinks channel boundary channel if non-zero

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_ch_attr(void *twa, int ch, int blank, int hmirror,
			int vmirror, int blink);

/*! \brief sets the recording mode

  \param twa pointer to the TWA2835 instance

  \param frame if non-zero enables frame mode

  \param dvr if non-zero enables DVR mode

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_mode(void *twa, int frame, int dvr);

/*! \brief freezes a channel on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_freeze_ch(void *twa, int ch);

/*! \brief puts a channel in continuos mode on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_continue_ch(void *twa, int ch);

/*! \brief updates a channel when freezed on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_strobe_ch(void *twa, int ch);

/*! \brief puts channel in switched mode on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param period (in number of fields) to wait befor hopping to the
  next channel

  \param ch_len lenght of the list of channel to be switched

  \param ch_list array with channel number to be displayed in sequence

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_switch_mode(void *twa, int ch, int period, int ch_len, int *ch_list);

/*! \brief changes a channel in switched mode to the given channel on the record path

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_switch_trigger(void *twa, int ch);

/*! \brief freezes the record path

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_freeze(void *twa);

/*! \brief puts the record path in continuos mode

  \param twa pointer to the TWA2835 instance

  \return 0 on success, otherwise non-zero
*/
int twa2835_rec_continue(void *twa);

/*! \brief adjust brightness on given input channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param v parameter value (0-255)

  \return 0 on success, otherwise non-zero
*/
int twa2835_brightness(void *twa, int ch, int v);

/*! \brief adjust contrast on given input channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param v parameter value (0-255)

  \return 0 on success, otherwise non-zero
*/
int twa2835_contrast(void *twa, int ch, int v);

/*! \brief adjust hue on given input channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param v parameter value (0-255)

  \return 0 on success, otherwise non-zero
*/
int twa2835_hue(void *twa, int ch, int v);

/*! \brief adjust saturation on given input channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param v parameter value (0-255)

  \return 0 on success, otherwise non-zero
*/
int twa2835_saturation(void *twa, int ch, int v);

/*! \brief returns status for an input channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param s struc sescribing status of given input channel

  \return 0 on success, otherwise non-zero
*/
int twa2835_get_status(void *twa, int ch, struct twa2835_status *s);

/*! \brief switches between A and B input for each channel

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param ab if zero selects A input, if 1 B

  \return 0 on success, otherwise non-zero
*/
int twa2835_ch_ab(void *twa, int ch, int ab);

/*! \brief returns the motion detection engine status

  \param twa pointer to the TWA2835 instance

  \param s outputs the status of various modules in the motion detection engine

  \return 0 on success, otherwise non-zero
*/
int twa2835_get_motion_status(void *twa, struct twa2835_motion_status *s);

/*! \brief sets which events will trigger an interrupt from the TWA2835

  \param twa pointer to the TWA2835 instance

  \param nv sets which channels will triegger a video loss event. Bits
  0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \param md sets which channels will triegger a motion detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \param bd sets which channels will triegger a blind detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \param nd Sets which channels will triegger a night detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \return 0 on success, otherwise non-zero
*/
int twa2835_enairq(void *twa, int nv, int md, int bd, int nd);

/*! \brief returns the status of various detection events that may
  have triggered an interrupt

  \param twa pointer to the TWA2835 instance

  \param nv returns which channels triggered a video loss event. Bits
  0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7 on mux B

  \param md returns which channels trieggered a motion detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \param bd returns which channels trieggered a blind detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \param nd returns which channels trieggered a night detection
  event. Bits 0-3 are channels 0-3 on mux A, bits 4-7 are channels 4-7
  on mux B

  \return 0 on success, otherwise non-zero
*/
int twa2835_get_irq_status(void *twa, int *nv, int *md, int *bd, int *nd);

/*! \brief sets the motion detection parameters

  \param twa pointer to the TWA2835 instance

  \param ch channel number (0-3)

  \param disable if 1 disables channel

  \param md_reffld controls the updating time of reference field for
  motion detection (0 every frame, 1 according to md_speed)

  \param bd_celsens the threshold of cell for blind detection (0-3)

  \param bd_lvsens the threshold of level for blind detection (0-15)

  \param nd_lvsens the threshold of level for night detection (0-15)

  \param nd_tmpsens the threshold of temporal sensitivity for night detection (0-15)

  \param md_fld 0 = odd field, 1 = even field 2 = any 3 = both

  \param md_align horizontal starting position for motion detection (0-15)

  \param md_celsens the threshold of sub-cell number for motion detection (0-3)

  \param md_dual_ena if 1 enable the non-realtime motion detection mode (8 channel mode)

  \param md_lvsens the level sensitivity of motion detector (0-31)

  \param md_speed velocity of motion detector (0-63)

  \param md_tmpsens temporal sensitivity of motion detecto (0-15)

  \param md_spsens spatial sensitivity of motion detecto (0-15)

  \param md_mask an array of 12 ints. Every int rapresents a
  row. Every bit in an int corresponds to a cell. If bit is one that
  cell is ignored for motion detection.

  \return 0 on success, otherwise non-zero
 */
int twa2835_set_motion_par(void *twa, int ch, int disable, int md_reffld,
			   int bd_celsens, int bd_lvsens, int nd_lvsens,
			   int nd_tmpsens, int md_fld, int md_align,
			   int md_celsens, int md_dual_ena, int md_lvsens,
			   int md_speed, int md_tmpsens, int md_spsens,
			   int *md_mask);

/*! \brief sets the audio part in simple pass through mode

  \param twa pointer to the TWA2835 instance

  \param in_gain gain for 4 input channel (0-15)

  \param mute can mute each of the input channel if 1

  \param out_gain output gain (0-15)

  \return 0 on success, otherwise non-zero
 */
int twa2835_audio_passtru(void *twa, int *in_gain, int *mute, int out_gain);

#endif
