/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.

  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

//#define DEBUG 1

//#define DEBUG_IO 1

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/device.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <linux/workqueue.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>
#include <asm/dma.h>

#include "plx9056.h"
#include "plx9056_registers.h"
#include "driver.h"
#include "adv202.h"
#include "twa2835_linux.h"
#include "version.h"

/* module parameters */

int images_max_size = 100*1024;
module_param(images_max_size, int, S_IRUGO);
int images_n = 10;
module_param(images_n, int, S_IRUGO);

int quiet = 0;
module_param(quiet, int, S_IRUGO);

int irq_force = -1;
module_param(irq_force, int, S_IRUGO);

int plx_major = 0;
module_param(plx_major, int, S_IRUGO);

int plx_minor = 0;
module_param(plx_minor, int, S_IRUGO);

/* for sysfs */
static struct class *plx_class;
static int ref;

/* kernel version specifics */
#define mem_map_reserve(adr)		SetPageReserved(adr)
#define mem_map_unreserve(adr)		ClearPageReserved(adr)

static inline unsigned long kvirt_to_pa(unsigned long adr)
{
   unsigned long kva;

   kva = (unsigned long) page_address(vmalloc_to_page((void *) adr));
   kva |= adr&(PAGE_SIZE-1);   /* restore the offset */
   return __pa(kva);
}
/* end kernel version specifics */

void plx_reg_write32(struct pci_dev *dev,u32 reg,u32 data)
{
	struct plx_dev *pdev = pci_get_drvdata(dev);

#ifdef DEBUG_IO
	printq("w %08x = %08x\n", reg, data);
#endif
	writel(data,pdev->baseaddress + reg);
}

u32 plx_reg_read32(struct pci_dev *dev,u32 reg)
{
	struct plx_dev *pdev = pci_get_drvdata(dev);
	u32 data;

	data = readl(pdev->baseaddress + reg);
#ifdef DEBUG_IO
	printq("r %08x = %08x\n", reg, data);
#endif
	return data;
}

void PLX9056_WriteConfigReg ( struct plx_dev * pdev, u32 reg, u32 val, int mode)
{
	switch (mode) {
	case SIZE_8BITS:
		writeb(val, pdev->iobase + reg);
		break;
	case SIZE_16BITS:
		writew(val, pdev->iobase + reg);
		break;
	case SIZE_32BITS:
		writel(val, pdev->iobase + reg);
		break;
	default:
		printq("BUG in %s, reg is %08x\n\n", __FUNCTION__, reg);
	}
}

u32 PLX9056_ReadConfigReg ( struct plx_dev * pdev, u32 reg, int mode)
{
	u32 r = 0;

	switch (mode) {
	case SIZE_8BITS:
		r = readb(pdev->iobase + reg);
		break;
	case SIZE_16BITS:
		r = readw(pdev->iobase + reg);
		break;
	case SIZE_32BITS:
		r = readl(pdev->iobase + reg);
		break;
	default:
		printq("BUG in %s, reg is %08x\n\n", __FUNCTION__, reg);
	}
	return r;
}

static void * rvmalloc(signed long size)
{
	struct page *page;
	void * mem;
	unsigned long adr;

	mem=vmalloc_32(size);
	if (NULL == mem)
		printq(KERN_INFO "bttv: vmalloc_32(%ld) failed\n",size);
	else {
		/* Clear the ram out, no junk to the user */
		memset(mem, 0, size);
	        adr=(unsigned long) mem;
		while (size > 0) {
			page = vmalloc_to_page((void *)adr);
			mem_map_reserve(page);
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
	}
	return mem;
}

static void rvfree(void * mem, signed long size)
{
	struct page *page;
        unsigned long adr;

	if (mem) {
	        adr=(unsigned long) mem;
		while (size > 0) {
			page = vmalloc_to_page((void *)adr);
			mem_map_unreserve(page);
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
		vfree(mem);
	}
}

int dma_map(struct plx_dma *dma, u8 *buf, int size, int from_device)
{
	int n, last, i;
	struct plx_dma_desc *pd = dma->pg_dma;
	struct page *page;

	n = size/PAGE_SIZE;
	last = size & (PAGE_SIZE-1);
	if (last)
		n++;

	for (i=0 ; i<n; i++) {
		page = vmalloc_to_page((void *) &buf[i * PAGE_SIZE]);
		pd[i].BufferPhysicalAddress = dma_map_page(&dma->plx->pci->dev,
							   page, 0, PAGE_SIZE,
							   from_device ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
		pd[i].BufferLocalAddress = 0x80 + 4 * 1;
		if (i == (n-1) && last)
			pd[i].BufferTransferSize = last;
		else
			pd[i].BufferTransferSize = PAGE_SIZE;
		pd[i].NextDescriptorPhysicalAddress = ((i + 1) * sizeof(struct plx_dma_desc) + dma->pg_dma_phys) |
			PLX_DESC_PTR_LOCATION_PCI | (from_device ? PLX_DESC_PTR_LCL_TO_PCI : 0);
	}
	pd[i].BufferPhysicalAddress = 0;
	pd[i].BufferLocalAddress = 0x80 + 4 * 1;
	pd[i].BufferTransferSize = 0;
	pd[i].NextDescriptorPhysicalAddress = PLX_DESC_PTR_END_OF_CHAIN |
		PLX_DESC_PTR_LOCATION_PCI | (from_device ? PLX_DESC_PTR_LCL_TO_PCI : 0);
	return 0;
}

int dma_unmap(struct plx_dma *dma, u8 *buf, int size, int from_device)
{
	int n, last, i;
	struct plx_dma_desc *pd = dma->pg_dma;

	n = size/PAGE_SIZE;
	last = size & (PAGE_SIZE-1);
	if (last)
		n++;

	for (i=0 ; i<n; i++) {
		dma_unmap_page(&dma->plx->pci->dev,
			       pd[i].BufferPhysicalAddress, PAGE_SIZE,
			       from_device ? DMA_FROM_DEVICE : DMA_TO_DEVICE);
	}
	return 0;
}


int alloc_global_memory(struct plx_dma *dma) {
	dma->vbuffer = rvmalloc(dma->img_n * dma->img_max_size);
	dma->head = dma->tail = 0;
	dma->overflow = 0;

	dma->img = kzalloc(sizeof(struct plx_image) * dma->img_n, GFP_KERNEL);
	dma->pg_dma = dma_alloc_coherent(&dma->plx->pci->dev, PAGE_SIZE,
					 &dma->pg_dma_phys, GFP_KERNEL);
	return 0;
}

void free_global_memory(struct plx_dma *dma) {
	rvfree(dma->vbuffer, dma->img_n * dma->img_max_size);
	dma_free_coherent(&dma->plx->pci->dev, PAGE_SIZE,
			  dma->pg_dma, dma->pg_dma_phys);
	kfree(dma->img);
}


void start_dma(struct plx_dma *dma, int ch)
{
	struct plx_dev *bSlave = dma->plx;
	unsigned long	FirstDescriptorPhyAddr;
	unsigned long  DMAMode =
		PLX_DMA_MODE_LCLBUS32     	|
		PLX_DMA_MODE_SG_ENA       	|
		PLX_DMA_MODE_INT_DONE_ENA 	|
		PLX_DMA_MODE_CONST_LCL_ADDR|
		PLX_DMA_MODE_CLEAR_COUNT |
	  		PLX_DMA_MODE_DEMAND       	|
		PLX_DMA_MODE_INT_PCI;

	DMAMode |= PLX_DMA_MODE_TA_RDY_ENA;
	FirstDescriptorPhyAddr =  dma->pg_dma_phys;
	FirstDescriptorPhyAddr =  FirstDescriptorPhyAddr |
		PLX_DESC_PTR_LOCATION_PCI | (ch==1 ? 0 : PLX_DESC_PTR_LCL_TO_PCI) ;

	switch (ch)
	{
	case 0:
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_MODE, DMAMode, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_XFR_SIZE, 0, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_DESC_POINTER, FirstDescriptorPhyAddr, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_LCL_ADDR, 0x80 + 1 * 4, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_CMDSTS, PLX_CMDSTS_ENABLE, SIZE_8BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_CMDSTS, PLX_CMDSTS_ENABLE | PLX_CMDSTS_START, SIZE_8BITS );
		break;

	case 1:
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_MODE, DMAMode, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_XFR_SIZE, 0, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_DESC_POINTER, FirstDescriptorPhyAddr, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_LCL_ADDR, 0, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_CMDSTS, PLX_CMDSTS_ENABLE, SIZE_8BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_CMDSTS, PLX_CMDSTS_ENABLE | PLX_CMDSTS_START, SIZE_8BITS );
		break;
	}
}

void stop_dma(struct plx_dma *dma, int DMAChannel)
{
	struct plx_dev *bSlave = dma->plx;
	unsigned char DMACmdSts;
	unsigned long IntCtrlStat;
	unsigned long j;

	DMACmdSts = (unsigned char)PLX9056_ReadConfigReg( bSlave,
							  DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
							  sizeof(unsigned char) );

	if ( ( DMACmdSts & PLX_CMDSTS_ENABLE ) &&
	     (!( DMACmdSts & PLX_CMDSTS_DONE ) ) )
	{
		//Mask DMA interrupt
		IntCtrlStat = PLX9056_ReadConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, SIZE_32BITS);
		IntCtrlStat &= (DMAChannel ? (~PLX_INTCTRLSTS_DMA1_INT_ENA) : (~PLX_INTCTRLSTS_DMA0_INT_ENA));
		PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, IntCtrlStat, SIZE_32BITS );
		//Stop and abort DMA
		PLX9056_WriteConfigReg( bSlave, DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
					PLX_CMDSTS_ABORT, SIZE_8BITS);
		//Wait for completion
		j = jiffies + HZ;
		do {
			DMACmdSts = (unsigned char)PLX9056_ReadConfigReg( bSlave,
									  DMAChannel ? PLX_ADDR_DMA1_CMDSTS :
									  PLX_ADDR_DMA0_CMDSTS,
									  SIZE_8BITS);
			schedule();
		} while (!(DMACmdSts & PLX_CMDSTS_DONE) && time_before(jiffies, j));
		if (!(DMACmdSts & PLX_CMDSTS_DONE)) {
			printq("%s error: DMACmdSts is %08x\n", __FUNCTION__, DMACmdSts);
		}
	}
	//Clear DMA interrupts
	PLX9056_WriteConfigReg( bSlave, DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
				 PLX_CMDSTS_CLRINT, SIZE_8BITS );
}

static void init_dma(struct plx_dma *dma) {
	struct plx_dev *bSlave = dma->plx;

	pr_debug("%s %d\n", __FUNCTION__, __LINE__);
	// enable local address space 0, offset 0
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_LAS0BA, PLX_LADR0_ENABLE,SIZE_32BITS );
	// disable local burst and prefetch
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_LBRD0, PLX_REGSTARTUP_LBRD0,SIZE_32BITS );
	// set DMA endianess for PLX
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_ENDIAN, PLX_REGSTARTUP_ENDIAN,
				SIZE_8BITS );
	// set DMA arbitration mode
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_DMA_MODE_ARB, PLX_REGSTARTUP_DMA_MODE_ARB, SIZE_32BITS );
	// Disable interrupts
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS,0, SIZE_32BITS );
	// Clear possibly pending interrupt flags
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS,
				PLX_INTCTRLSTS_PM_INT_ACTIVE|PLX_INTCTRLSTS_PE_INT_ACTIVE, SIZE_32BITS );

#if 0
	/* dump registers */
	{
	  int i;

	  for(i=0; i < 0x200; i+=4) {
	    printk("plx9056 dump %08x=%08x\n", i, PLX9056_ReadConfigReg ( bSlave, i, SIZE_32BITS ));
	    schedule();
	  }
	}
#endif
}

/* file operations */

static void plx_vma_open(struct vm_area_struct *vma)
{
	struct plx_dma *pdev = (struct plx_dma *) vma->vm_private_data;

	pr_debug("VMA open, virt %lx, phys %lx\n",
		 vma->vm_start, vma->vm_pgoff << PAGE_SHIFT);
	pdev->is_mapped = 1;
}

static void plx_vma_close(struct vm_area_struct *vma)
{
	struct plx_dma *pdev = (struct plx_dma *) vma->vm_private_data;
	pr_debug("VMA close.\n");
	pdev->is_mapped = 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,25)

static struct page *plx_vma_nopage(struct vm_area_struct *vma,
				   unsigned long address, int *type)
{
	struct plx_dma *pdev = (struct plx_dma *) vma->vm_private_data;
	struct page *pageptr;
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long virtaddr = address - vma->vm_start + offset;
	unsigned long pageframe;

	//pr_debug("nopage %08lx (from %08lx) ", virtaddr, address);
	while(virtaddr >= (pdev->img_max_size * pdev->img_n))
		virtaddr -= (pdev->img_max_size * pdev->img_n);
	//pr_debug("-> %08lx\n", (unsigned long) &pdev->vbuffer[virtaddr]);

	pageframe = kvirt_to_pa((unsigned long) &pdev->vbuffer[virtaddr]) >> PAGE_SHIFT;
	if (!pfn_valid(pageframe))
		return NOPAGE_SIGBUS;
	pageptr = pfn_to_page(pageframe);
	get_page(pageptr);
	if (type)
		*type = VM_FAULT_MINOR;
	return pageptr;
}

static struct vm_operations_struct plx_nopage_vm_ops = {
	.open =   plx_vma_open,
	.close =  plx_vma_close,
	.nopage = plx_vma_nopage,
};

#else

static int plx_vma_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
{
	struct plx_dma *pdev = (struct plx_dma *) vma->vm_private_data;
	struct page *pageptr;
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long virtaddr = ((unsigned long)vmf->virtual_address) -
		vma->vm_start + offset;
	unsigned long pageframe;

	while(virtaddr >= (pdev->img_max_size * pdev->img_n))
		virtaddr -= (pdev->img_max_size * pdev->img_n);

	pageframe = kvirt_to_pa((unsigned long) &pdev->vbuffer[virtaddr]) >> PAGE_SHIFT;
	if (!pfn_valid(pageframe))
		return VM_FAULT_SIGBUS;
	pageptr = pfn_to_page(pageframe);
	get_page(pageptr);
	vmf->page = pageptr;
	return 0;
}

static struct vm_operations_struct plx_nopage_vm_ops = {
	.open =   plx_vma_open,
	.close =  plx_vma_close,
	.fault =  plx_vma_fault,
};

#endif

static int plx_mmap(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

	if (offset >= __pa(high_memory) || (filp->f_flags & O_SYNC))
		vma->vm_flags |= VM_IO;
	vma->vm_flags |= VM_RESERVED;

	vma->vm_ops = &plx_nopage_vm_ops;
	vma->vm_private_data = filp->private_data;
	plx_vma_open(vma);
	return 0;
}



static ssize_t plx_read(struct file *filp, char __user *buf, size_t count,
                loff_t *f_pos)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_image *img;
	int ret;
	unsigned long timeout = image_timeout(pdev);

	//dev_dbg(&dev->dev, "%s %d\n", __FUNCTION__, __LINE__);

	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {
		if (count < sizeof(struct ethjpg2k_linux_hdr)) {
			dev_dbg(&pdev->plx->pci->dev, "read encode size\n");
			return -EINVAL;
		}

		if (pdev->head == pdev->tail) {
			ret = wait_event_interruptible_timeout(pdev->pready, (pdev->head != pdev->tail),
							       timeout);
			if (ret == 0) {
				printq("timeout waiting for image\n");
				return -ETIMEDOUT;
			}
			else {
				if (ret == -ERESTARTSYS)
					return ret;
			}
		}


		img = &pdev->img[pdev->tail];
		pr_debug("reading slot %d: %08lx %8ld %08lx %ld\n", pdev->tail, img->hdr.field,
			 img->hdr.index, img->hdr.Info_u.info, img->hdr.size);
		BUG_ON(img->status != IMAGE_READY);

		img->hdr.start = pdev->tail * pdev->img_max_size;
		img->hdr.overflow = pdev->overflow;
		pdev->overflow = 0;
		img->hdr.ticket = pdev->tail;
		img->status = IMAGE_USER;

		if (copy_to_user(buf, &img->hdr, sizeof(struct ethjpg2k_linux_hdr)))
			return -EFAULT;
		return sizeof(struct ethjpg2k_linux_hdr);
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT) {
		struct ethjpg2k_linux_decode d;

		if (count < sizeof(struct ethjpg2k_linux_decode)) {
			dev_dbg(&pdev->plx->pci->dev, "read decode size\n");
			return -EINVAL;
		}

		if ((pdev->head + 1) % pdev->img_n == pdev->tail) {
			ret = wait_event_interruptible_timeout(pdev->pready,
							       ((pdev->head + 1) % pdev->img_n != pdev->tail),
							       timeout);
			if (ret == 0)
				return -ETIMEDOUT;
			else {
				if (ret == -ERESTARTSYS)
					return ret;
			}
		}
		d.size = pdev->img_max_size;
		d.start = pdev->head * pdev->img_max_size;
		img = &pdev->img[pdev->head];
		BUG_ON(img->status != IMAGE_IDLE);
		img->status = IMAGE_USER;
		dev_dbg(&pdev->plx->pci->dev, "free from %ld and %ld bytes\n", d.start, d.size);
		if (copy_to_user(buf, &d, sizeof(d)))
			return -EFAULT;
		return sizeof(d);
	}
	else {
		dev_dbg(&pdev->plx->pci->dev, "bad direction\n");
	}

	return -EBADFD;
}

static ssize_t plx_write(struct file *filp, const char __user *buf, size_t count,
                loff_t *f_pos)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_image *img;
	unsigned int ticket;

	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {

		if (count != sizeof(unsigned int)) {
			dev_dbg(&pdev->plx->pci->dev, "write encode size\n");
			return -EINVAL;
		}

		if (copy_from_user(&ticket, buf, sizeof(unsigned int)))
			return -EFAULT;

		pr_debug("write for ticket %d\n", ticket);

		if (pdev->tail != ticket) {
			printq("BUG @ %s %d: userspace returning bad ticket!\n",
			       __FUNCTION__, __LINE__);
			return -EBADMSG;
		}

		img = &pdev->img[ticket];
		BUG_ON(img->status != IMAGE_USER);
		img->status = IMAGE_IDLE;
		pdev->tail = (pdev->tail + 1) % pdev->img_n;
		return count;
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT){
		if (count != sizeof(struct ethjpg2k_linux_decode)) {
			dev_dbg(&pdev->plx->pci->dev, "write decode wsize\n");
			return -EINVAL;
		}
		img = &pdev->img[pdev->head];
		BUG_ON(img->status != IMAGE_USER);
		if (copy_from_user(&img->dec, buf, sizeof(struct ethjpg2k_linux_decode)))
			return -EFAULT;
		img->status = IMAGE_READY;
		pr_debug("filled slot %d @ %lu\n", pdev->head, jiffies);
		pdev->head = (pdev->head + 1) % pdev->img_n;
		wake_up_interruptible(&pdev->pready1);
		return count;
	}
	else {
		dev_dbg(&pdev->plx->pci->dev, "bad direction\n");
	}
	return -EBADFD;
}

static int plx_release(struct inode *inode, struct file *filp)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_dev *bSlave = pdev->plx;
	dev_dbg(&pdev->plx->pci->dev, "%s %d\n", __FUNCTION__, __LINE__);

	if (pdev->opened & PLX_RUNNING) {
		stop_encdec(pdev);
	}

	free_global_memory(pdev);

	bSlave->mode = PLX_NONE;
	pdev->opened = PLX_CLOSED;
	return 0;
}

static int plx_open(struct inode *inode, struct file *filp)
{
	struct plx_dma *pdev;
	struct plx_dev *bSlave;

	pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, iminor(inode));

	pdev = container_of(inode->i_cdev, struct plx_dma, cdev);
	bSlave = pdev->plx;

	if (pdev->opened != PLX_CLOSED)
		return -EEXIST;

	/*
	   mode              minor 0 (enc)                            minor 1 (dec)
	   PLX_ENCODE        PLX_OPENED_IN (outputs encoded data)     none
	   PLX_DECODE        none                                     PLX_OPENED_OUT (input encoded data)
	   PLX_HIPI_ENCODE   PLX_OPENED_IN (outputs encoded data)     PLX_OPENED_OUT (input raw data)
	   PLX_HIPI_DECODE   PLX_OPENED_IN (outputs raw data)         PLX_OPENED_OUT (input encode data)
	   PLX_LOOP          PLX_OPENED_IN (outputs raw data)         PLX_OPENED_OUT (input raw data)

	   DMA: channel 0 tied to minor 0 where data goes from ADV202 to CPU so plx_dev->dma[0]
	        channel 1 tied to minor 1 where data goes from CPU to ADV202 so plx_dev->dma[1]
	*/

	if (iminor(inode) & 1) {
		pdev->opened = PLX_OPENED_OUT;
		bSlave->mode = PLX_DECODE;
	}
	else {
		pdev->opened = PLX_OPENED_IN;
		bSlave->mode = PLX_ENCODE;
	}
	pr_debug("%s pdev %p opened %08x mode %08x\n", __FUNCTION__,
		 pdev, pdev->opened, bSlave->mode);

	alloc_global_memory(pdev);
	filp->private_data = pdev;

	return 0;
}

struct file_operations plx_fops = {
	.owner =    THIS_MODULE,
	.read =     plx_read,
	.write =    plx_write,
	.open =     plx_open,
	.release =  plx_release,
	.ioctl =    plx_ioctl,
	.mmap =	    plx_mmap,
};

static int plx_setup_cdev(struct cdev  *cdev, int index, struct file_operations *ops,
			  struct class *class, struct device *dev)
{
	int err = 0, devno = MKDEV(plx_major, index);
	char b[20];

	if (index > 127) {
		sprintf(b, "ctr1474-%dctrl", index-128);
	}
	else {
		if (index & 1) {
			sprintf(b, "ctr1474-%ddec", index/2);
		}
		else {
			sprintf(b, "ctr1474-%denc", index/2);
		}
	}
	cdev_init(cdev, ops);
	cdev->owner = THIS_MODULE;
	cdev->ops = ops;
	err = cdev_add (cdev, devno, 1);
	/* Fail gracefully if need be */
	if (err)
		printq("Error %d adding index %d", err, index);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
	class_device_create(class, NULL, devno, dev, b);
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,27)
	device_create(class, NULL, devno, dev, b);
#else
	device_create_drvdata(class, NULL, devno, dev, b);
#endif
#endif
	return err;
}

static int __devinit
plx_pci_probe (struct pci_dev *dev, const struct pci_device_id *id)
{
	struct plx_dev  *pdev;
	int status = 1;
	dev_t devn = 0;
	dev_t devn1 = 0;
	int i;
	PETHJPG2K pethjpg2k;

	dev_dbg(&dev->dev,"%s %d\n", __FUNCTION__, __LINE__);

	if (ref == 0) {
		plx_class = class_create(THIS_MODULE, "ctr1474");
	}
	ref += 1;

	if (pci_enable_device(dev) < 0){
		printq("failed in enabing the device\n");
		return -ENODEV;
	}
	if (!dev->irq) {
		if (irq_force)
			dev->irq = irq_force;
		else {
			printq("found plx device with no IRQ assigned.");
			printq("check BIOS settings!");
			return -ENODEV;
		}
	}

	pdev = kzalloc(sizeof(*pdev), GFP_KERNEL);
	init_waitqueue_head(&pdev->dma[0].pready);
	init_waitqueue_head(&pdev->dma[1].pready);
	init_waitqueue_head(&pdev->dma[0].pready1);
	init_waitqueue_head(&pdev->dma[1].pready1);
	init_completion(&pdev->dma[0].dma_done);
	init_completion(&pdev->dma[1].dma_done);
	pdev->dma[0].plx = pdev;
	pdev->dma[1].plx = pdev;
	pdev->pci = dev;
	pdev->dma[0].img_max_size = images_max_size;
	pdev->dma[0].img_n = images_n;
	pdev->dma[1].img_max_size = images_max_size;
	pdev->dma[1].img_n = images_n;
	sprintf(pdev->name, "plx%d", plx_minor);

	/* Grab the PLX PCI mem maped port start address we need  */
	pdev->pci_io_base = pci_resource_start(dev, 0);
	dev_dbg(&dev->dev, "pci io base %08x, irq %d\n", pdev->pci_io_base, dev->irq);;

	pdev->iolength = pci_resource_len(dev, 0);
	dev_dbg(&dev->dev, "pci io length %d\n", pdev->iolength);
	if(!request_mem_region(pdev->pci_io_base, pdev->iolength, pdev->name)){
		printq("host controller already in use!\n");
		status = -EBUSY;
		goto err0;
	}

	pdev->iobase = ioremap_nocache(pdev->pci_io_base, pdev->iolength);
	if(!pdev->iobase){
		printq("can not map io memory to system memory\n");
		status = -ENOMEM;
		goto err1;
	}

	/* Grab the PLX PCI shared memory */
	pdev->pci_mem_phy0 = pci_resource_start (dev, 2);
	dev_dbg(&dev->dev, "pci base address = %x\n", pdev->pci_mem_phy0);
	pdev->length = pci_resource_len(dev,2);

	if(check_mem_region(pdev->pci_mem_phy0, pdev->length)<0){
		printq("host controller already in use\n");
		status = -EBUSY;
		goto err2;
	}

	if(!request_mem_region(pdev->pci_mem_phy0, pdev->length, pdev->name)){
		printq("host controller already in use\n");
		status =  -EBUSY;
		goto err2;
	}

	/*map available memory*/
	pdev->baseaddress = ioremap_nocache(pdev->pci_mem_phy0,pdev->length);
	if(!pdev->baseaddress){
		printq("memory map problem\n");
		status =  -ENOMEM;
		goto err3;
	}
	pci_set_drvdata (dev, pdev);
	plx_request_irq(pdev->pci, pdev->pci);
	init_dma(&pdev->dma[0]);

	pci_set_master(dev);

	if (plx_major) {
		devn = MKDEV(plx_major, plx_minor);
		status = register_chrdev_region(devn, 2, "plx");
	} else {
		status = alloc_chrdev_region(&devn, plx_minor, 2,
					     "plx");
		plx_major = MAJOR(devn);
	}
	if (status < 0) {
		printq("plx: can't get major %d\n", plx_major);
		goto err4;
	}

	for(i=0;i<2;i++) {
		dev_dbg(&dev->dev, "registering minor %d\n", plx_minor);
		pdev->dma[i].minor = plx_minor;
		plx_setup_cdev(&pdev->dma[i].cdev, plx_minor, &plx_fops,
			       plx_class, &dev->dev);
		dev_dbg(&dev->dev, "%p %p\n", pdev, &pdev->dma[i].cdev);
		plx_minor ++;
	}

	dev_dbg(&dev->dev, "registering twa minor %d\n", plx_minor/2 + 127);
	devn1 = MKDEV(plx_major, plx_minor/2 + 127);
	register_chrdev_region(devn1, 1, "plx");
	plx_setup_cdev(&pdev->cdev, plx_minor/2 + 127, &twa2835_fops,
		       plx_class, &dev->dev);

	pethjpg2k = kzalloc(sizeof(*pethjpg2k), GFP_KERNEL);
	pdev->pethjpg2k = pethjpg2k;
	pethjpg2k->plx = pdev;
	pethjpg2k->PLX9056_BAR2 = (unsigned long) pdev->baseaddress;

	init_waitqueue_head(&pdev->twa_wait);
	init_cpld(pethjpg2k);

	return 1;

        // err5:
	//unregister_chrdev_region(devn, 2);

 err4:
	iounmap(pdev->baseaddress);
 err3:
	release_mem_region(pdev->pci_mem_phy0, pdev->length);
 err2:
	iounmap(pdev->iobase);
 err1:
	release_mem_region(pdev->pci_io_base, pdev->iolength);
 err0:
	kfree(pdev);
	return status;
}

static void __devexit
plx_pci_remove (struct pci_dev *dev)
{
	struct plx_dev *pdev = pci_get_drvdata(dev);
	dev_t devn = MKDEV(plx_major, pdev->dma[0].minor);
	dev_t devn1 = MKDEV(plx_major, pdev->dma[1].minor);
	dev_t devn2 = MKDEV(plx_major, (pdev->dma[1].minor + 1) / 2 + 127);

	device_destroy(plx_class, devn);
	cdev_del(&pdev->dma[0].cdev);
	device_destroy(plx_class, devn1);
	cdev_del(&pdev->dma[1].cdev);
	unregister_chrdev_region(devn, 2);
	device_destroy(plx_class, devn2);
	cdev_del(&pdev->cdev);
	unregister_chrdev_region(devn2, 1);

	ref -= 1;
	if (ref == 0) {
		class_destroy(plx_class);
	}

	plx_free_irq(dev, dev);

	release_mem_region(pdev->pci_io_base, pdev->iolength);
	release_mem_region(pdev->pci_mem_phy0, pdev->length);
	iounmap(pdev->baseaddress);
	iounmap(pdev->iobase);

	kfree(pdev->pethjpg2k);
	kfree(pdev);

	return;
}


static const struct pci_device_id __devinitdata plx_pci_ids [] = {
	{
		/* handle PCI BRIDGE  manufactured by PLX */
		class:          ((PCI_CLASS_BRIDGE_OTHER << 8) | (0x06 << 16)),
		/* TODO: Eurotech vendorid / productid */
#ifdef ANALOG_DEMO
		class_mask:     0,
		vendor:         0x11d4,
		device:         0x0202,
#else
		class_mask:     ~0,
		vendor:         0x10B5,
		device:         0x9056,
#endif
		subvendor:      PCI_ANY_ID,
		subdevice:      PCI_ANY_ID,
	},
	{ /* end: all zeroes */ }
};

MODULE_DEVICE_TABLE (pci, plx_pci_ids);

/* Pci driver interface functions */
static struct pci_driver plx_pci_driver = {
	name:           "plx9056",
	id_table:       &plx_pci_ids[0],
	probe:         plx_pci_probe,
	remove:        plx_pci_remove,
};

static int __init plx_module_init(void)
{
	int result = 0;
	printk("Eurotech CTR1474 driver, version " VERSION " compiled on " __DATE__ " " __TIME__ "\n");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,21)
	if((result = pci_register_driver(&plx_pci_driver)) < 0) {
#else
	if((result = pci_module_init(&plx_pci_driver)) < 0) {
#endif
		printq("PCI Iinitialization Fail(error = %d)\n",result);
		return result;
	}
	else
		pr_debug("%s PCI Initialization Success \n",plx_pci_driver.name);
	return result;
}

static void __exit plx_module_cleanup (void)
{
	pr_debug("%s: %d\n",__FUNCTION__, __LINE__);

	pci_unregister_driver (&plx_pci_driver);
}


MODULE_AUTHOR ("Christian Pellegrin <chri@evolware.org>");
MODULE_DESCRIPTION("Eurotech CTR1474 driver, version " VERSION " compiled on " __DATE__ " " __TIME__);
MODULE_LICENSE ("GPL");

module_init (plx_module_init);
module_exit (plx_module_cleanup);


