/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.

  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <time.h>

#include "libctr1474.h"
#include "twa2835.h"

#include "twa2835.c"

int simple_ctr1474_gpio_read(struct ctr1474_ch* e) {
  int fd = e->ctrl_fd;
  int t[1];
  int ret;

  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_GPIO_READ, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_GPIO_READ");
  }
  return t[0];
}

int adv212_read_reg(struct ctr1474_ch *e, int addr) {
  int fd = e->ctrl_fd;
  int t[2];
  int ret;

  t[0] = addr;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_ADV212_READ_REG, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_ADV212_READ_REG");
  }
  return t[1];
}

void adv212_write_reg(struct ctr1474_ch *e, int addr, int data) {
  int fd = e->ctrl_fd;
  int t[2];
  int ret;

  t[0] = addr;
  t[1] = data;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_ADV212_WRITE_REG, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_ADV212_WRITE_REG");
  }
}

int twa2835_read_reg(void *twa, int addr) {
  int fd = (long) twa;
  int t[2];
  int ret;

  t[0] = addr;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_TW2835_READ_REG, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_TW2835_READ_REG");
  }
  return t[1];
}

void twa2835_write_reg(void *twa, int addr, int data) {
  int fd = (long) twa;
  int t[2];
  int ret;

  t[0] = addr;
  t[1] = data;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_TW2835_WRITE_REG, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_TW2835_WRITE_REG");
  }
}

int twa2835_batch(void *twa, int *batch) {
  int fd = (long) twa;
  int ret;

  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_TW2835_BATCH, batch)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_TW2835_BATCH");
  }
  return ret;
}

void twa2835_reset(void *twa) {
  int fd = (long) twa;
  int ret;

  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_TW2835_RESET, NULL)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_TW2835_RESET");
  }
}

unsigned long twa2835_get_sec(void) {
  return time(NULL);
}

int twa2835_wait_motion(void *twa, int sec, int *nv, int *md, int *bd, int *nd) {
  int fd = (long) twa;
  int t[1];
  int ret;

  t[0] = sec > 1 ? sec : 1;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_TW2835_TOUT_SEC, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_TW2835_TOUT_SEC");
  }

  read(fd, t, sizeof(t[0]));
  usleep(100 * 1000);	/* wait a field */
  twa2835_get_irq_status(twa, nv, md, bd, nd);
  return 0;
}


struct ctr1474_ch* simple_ctr1474(char *dev, char *cdev) {
  struct ctr1474_ch *e;
  void *m;
  int ret;

  e = malloc(sizeof(*e));
  assert(e);
  memset(e, 0, sizeof(*e));

  e->last_isNtsc = 666;

  e->fd = open(dev, O_RDWR);
  if (e->fd == -1) {
    perror("simple_ctr1474 open");
    free(e);
    return NULL;
  }

  e->ctrl_fd = open(cdev, O_RDWR);
  if (e->ctrl_fd == -1) {
    perror("simple_ctr1474 ctrl open");
    free(e);
    return NULL;
  }

  if ((ret = ioctl(e->fd, IOCTL_ETHJPG2K_LINUX_GETBSIZE, &e->bsize)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_LINUX_GETBSIZE");
    free(e);
    return NULL;
  }
  e->bsize *= 4096;
  //printf("buffer size: %ld\n", e->bsize);

  m = mmap(NULL, e->bsize, PROT_READ|PROT_WRITE , MAP_SHARED, e->fd, 0);
  if (m == MAP_FAILED) {
    perror("mmap");
    free(e);
    return NULL;
  }
  e->b = (unsigned char *) m;
  memset(m, 11, e->bsize);	/* pagefault memory */

  e->bsize /= 4;

  return e;
}


int simple_ctr1474_board_init(struct ctr1474_ch *e, int isNtsc , char *b,
			    ETHJPG2K_FIRMWARE_DATA *fwdata,
			    ETHJPG2K_FIRMWARE_PARAMS *fwparams) {
  int ret;
  unsigned char firmware[32768];

  FILE *f;

  f = fopen(b, "r");
  if (!f) {
    perror("open firmware");
    fprintf(stderr, "cannot open firmware: %s\n", b);
    return -1;
  }
  fread(firmware, 32768, 1, f);
  fclose(f);

  fwdata->pFirmwareParameters = fwparams;
  fwdata->pFirmware = (unsigned long *) firmware;

  if ((ret = ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE, fwdata)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE");
    return ret;
  }

  if (1 /*isNtsc != e->last_isNtsc*/) {
    if (e->initialized) {
      twa2835_release(TWA(e));
    }
    if ((ret = twa2835_init(TWA(e), isNtsc)) < 0) {
      fprintf(stderr, "twa2835_init error\n");
      return ret;
    }
    e->initialized = 1;
    if ((ret = ioctl(e->fd, IOCTL_ETHJPG2K_LINUX_INIT, isNtsc)) == -1) {
      perror("ioctl IOCTL_ETHJPG2K_LINUX_INIT");
      return ret;
    }
    e->last_isNtsc = isNtsc;
  }

  e->cod_style = fwparams->DecodeParams.COD_STYLE;
  e->video_norm = fwparams->DecodeParams.VFORMAT;

  return 0;
}


int simple_ctr1474_encoder_run(struct ctr1474_ch* e, ctr1474_frame_cb cb) {
  int ret;
  struct ethjpg2k_linux_hdr hdr;


  if ((ret = ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_START");
    free(e);
    return ret;
  }

  ret = 0;
  while(!ret) {
    ret = read(e->fd, &hdr, sizeof(hdr));
    if (ret < 0) {
      perror("read hdr");
      break;
    }
    ret = cb(&hdr, e->b + hdr.start, hdr.size*4);
    write(e->fd, &hdr.ticket, sizeof(unsigned long));
  }
  ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

int simple_ctr1474_decoder_run(struct ctr1474_ch* e, ctr1474_len_cb cbl, ctr1474_frame_in_cb cbf, int add_hdr) {
  int ret, imsize, hdr_size = 0;
  int count = 0, j, field;
  struct ethjpg2k_linux_decode dec;

  if (add_hdr)
    hdr_size = 16;

  if ((ret = ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }

  while(1) {
    ret = imsize = cbl();
    if (ret < 0)
      break;
    while (1) {
      ret = read(e->fd, &dec, sizeof(dec));
      if (ret < 0) {
	perror("decoder read");
	goto out_of_here;
      } else {
	if (imsize + hdr_size < dec.size)
	  break;
	usleep(10 * 1000);
      }
    }
    cbf(e->b + dec.start + hdr_size, imsize, &field);
    dec.size = imsize + hdr_size;
    if (add_hdr) {
      unsigned int *h = (unsigned int *) (e->b + dec.start);

      h[0] = 0xfffffff0 | field;
      h[1] = count;
      h[2] = (e->cod_style << 24) | (e->video_norm << 16) | 2;
      h[3] = imsize / 4;
      for(j=0; j < 4; j++) {
	h[j] = htonl(h[j]);
      }
    }
    ret = write(e->fd, &dec, sizeof(dec));
    if (ret < 0) {
      perror("decoder write");
      break;
    }
    count ++;
  }
 out_of_here:
  ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

void simple_ctr1474_force_stop(struct ctr1474_ch* e) {
  ioctl(e->fd, IOCTL_ETHJPG2K_ENCDEC_STOP);
}

void simple_ctr1474_free(struct ctr1474_ch* e) {
  close(e->fd);
  free(e);
}

void simple_ctr1474_board_noinit(struct ctr1474_ch *e, int isNtsc) {
  e->last_isNtsc = isNtsc;
}

int simple_ctr1474_hipi(struct ctr1474_ch *enc, struct ctr1474_ch *dec,
		      char *fw, ETHJPG2K_FIRMWARE_DATA *fwdata, ETHJPG2K_FIRMWARE_PARAMS *fwparams,
		      int n,
		      char **in, int *ins,
		      char **out, int *outs) {
  int ret=0;
  int i,j;
  struct ethjpg2k_linux_decode dech;
  struct ethjpg2k_linux_hdr hdr;
  unsigned char firmware[32768];
  FILE *f;
  int ns = 2;			/* this accounts for HIPI being one frame behind */

  f = fopen(fw, "r");
  if (!f) {
    perror("open firmware");
    fprintf(stderr, "cannot open firmware: %s\n", fw);
    return -1;
  }
  fread(firmware, 32768, 1, f);
  fclose(f);

  fwdata->pFirmwareParameters = fwparams;
  fwdata->pFirmware = (unsigned long *) firmware;
  if ((ret = ioctl(enc->fd, IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE, fwdata)) == -1) {
    perror("ioctl IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE");
    return ret;
  }

  if ((ret = ioctl(dec->fd, IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl dec IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }
  if ((ret = ioctl(enc->fd, IOCTL_ETHJPG2K_ENCDEC_START)) == -1) {
    perror("ioctl enc IOCTL_ETHJPG2K_ENCDEC_START");
    return ret;
  }

  for(i=0; i<n;i++) {
    for (j=0; j<ns; j++) {
      while (1) {
	ret = read(dec->fd, &dech, sizeof(dech));
	if (ret < 0) {
	  perror("decoder read");
	} else {
	  if (ins[i] < dech.size)
	    break;
	  usleep(100 * 1000);
	}
      }
      memcpy(&dec->b[dech.start], in[i], ins[i]);
      dech.size = ins[i];
      ret = write(dec->fd, &dech, sizeof(dech));
      if (ret < 0) {
	perror("decoder write");
	break;
      }
    }
    ns = 1;

    ret = read(enc->fd, &hdr, sizeof(hdr));
    if (ret < 0) {
      perror("encoder read hdr");
      break;
    }
    out[i] = malloc(hdr.size*4);
    assert(out[i]);
    outs[i] = hdr.size*4;
    memcpy(out[i], &enc->b[hdr.start], hdr.size*4);
    write(enc->fd, &hdr.ticket, sizeof(unsigned long));
  }
  ioctl(enc->fd, IOCTL_ETHJPG2K_ENCDEC_STOP);
  ioctl(dec->fd, IOCTL_ETHJPG2K_ENCDEC_STOP);
  return ret;
}

static int simple_ctr1474_gpio_write_internal(struct ctr1474_ch* e, int val, int max, int shift) {
  int fd = e->ctrl_fd;
  int t[1];
  int ret;

  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_GPIO_WRITE_READ, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_GPIO_WRITE_READ");
    return -1;
  }
  t[0] &= ~(max << shift);
  t[0] |= val << shift;
  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_GPIO_WRITE, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_GPIO_WRITE");
    return -1;
  }

  return 0;
}

int simple_ctr1474_pb(struct ctr1474_ch* e, int enable) {
  return simple_ctr1474_gpio_write_internal(e, enable != 0, 1, 9);
}

int simple_ctr1474_gpio_write(struct ctr1474_ch* e, int data) {
  if (data < 0)
    data = 0;
  if (data > 0x1ff)
    data = 0x1ff;
  return simple_ctr1474_gpio_write_internal(e, data, 0x1ff, 0);
}

int simple_ctr1474_led(struct ctr1474_ch* e, int data) {
  if (data < 0)
    data = 0;
  if (data > 3)
    data = 3;
  return simple_ctr1474_gpio_write_internal(e, data, 3, 10);
}

int simple_ctr1474_encoder_quality(struct ctr1474_ch* e, int q) {
  int fd = e->fd;
  int t[1] = {q};
  int ret;

  if ((ret = ioctl(fd, IOCTL_ETHJPG2K_ENC_QUALITY, t)) < 0) {
    perror("ioctl IOCTL_ETHJPG2K_ENC_QUALITY");
    return -1;
  }
  return 0;
}
