/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.

  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#ifndef _LIBCTR1474_H_
#define _LIBCTR1474_H_  1

#include "ethjpg2k.h"
#include "adv202_hdr.h"
#include "twa2835.h"

/*! \brief ctr1474 instance

  The struct rapresents an instance of a jpeg2000 encoder or
  decorder. It should be an opaque object for the application.

 */
struct ctr1474_ch {
  int fd; 			/*!< encoding/decoding device file descriptor  */
  long ctrl_fd;			/*!< control device file descriptor */

  unsigned char * b;		/*!< pointer to the buffer used to communicate with the CTR1474 */
  unsigned long bsize;		/*!< size of the communication buffer */
  int last_isNtsc;		/*!< non-zero if currently in NTSC mode */
  int cod_style;		/*!< current coding style */
  int video_norm;		/*!< current video norm */
  int initialized;		/*!< non-zero if already initialized */
};


/*! \brief returns the twa2835 handler for the given ctr1474 instance
 */
#define TWA(p) ( (void *) p->ctrl_fd )

/*! \brief callback for the jpeg2000 encoder

  User should provide an implementation of a function with this
  signature. Every time the ctr1474 acquires a frame the provided
  function will be called.

  \param hdr the header that describes the captured image

  \param data the pointer to the image bits (jpeg2000 encoded)

  \param len the length (in bytes) of the image

  \return 0 to continue acquiring, 1 to stop it

 */
typedef int (*ctr1474_frame_cb) (struct ethjpg2k_linux_hdr *hdr, unsigned char * data, unsigned long len);

/*! \brief callback for the jpeg2000 decoder

  User should provide an implementation of a function with this
  signature. Every time the ctr1474 needs another image to decode this
  function is called.

  \param data where the user should but the image bits

  \param len the maximum length for the image

  \param field 0/1 for even/odd field when deinterlacing

  \return 0 to continue decoding, 1 to stop it

*/
typedef void (*ctr1474_frame_in_cb) (unsigned char * data, unsigned long len, int *field);

/*! \brief callback for the jpeg2000 decoder

  User should provide an implementation of a function with this
  signature. Every time the ctr1474 needs to know how much space is
  needed for the next  image to be docede this function is called.

  \return length, in bytes, of the next jpeg2000 image to be decoded

*/
typedef int (*ctr1474_len_cb) (void);

/*! \brief creates an jpeg2000 encoder/decoder instance

  This is the first function called by the application when accessing
  the ctr1474. It connects to the device driver.

  \param dev the encoding or decoding /dev entry

  \param cdev the control /dev entry

  \return the ctr1474 instance, NULL on error

*/
struct ctr1474_ch* simple_ctr1474(char *dev, char *cdev);

/*! \brief initializes the ctr1474

  \param e the ctr1474 instance

  \param isNtsc if non-zero sets NTSC mode.

  \param b filename of the firmware for the ADV212

  \param fwdata firmware description

  \param fwparams parameters for the firmware

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_board_init(struct ctr1474_ch *e, int isNtsc, char *b,
			    ETHJPG2K_FIRMWARE_DATA *fwdata,
			    ETHJPG2K_FIRMWARE_PARAMS *fwparams);

/*! \brief setup the ctr1474 instance pointer without actually initializing the hardware

  This function should be called before starting the encoding/decoding
  process if we don't want to alter the setting of the video front-end.

  \param e the ctr1474 instance

  \param isNtsc if non-zero sets NTSC mode

  \return 0 on success, otherwise non-zero
*/
void simple_ctr1474_board_noinit(struct ctr1474_ch *e, int isNtsc);

/*! \brief starts the encoding process

  \param e the ctr1474 instance

  \param cb the user supplied callback which is called ater every acquired image

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_encoder_run(struct ctr1474_ch* e, ctr1474_frame_cb cb);

/*! \brief set the quality of output stream for a running encoder

  \param e the ctr1474 instance

  \param q the quality parameter (RCVAL)

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_encoder_quality(struct ctr1474_ch* e, int q);

/*! \brief starts the decoding process

  \param e the ctr1474 instance

  \param cbl the user supplied callback which is called to get the
  size of the next image to be decoded

  \param cbf the user supplied callback which is called to get the
  bits for the next image to be decoded

  \param ad_hdr if 1 an image header needed for the ADV212 is
  automatically generated, otherwise it must be supplied by the user
  before the bits of the actual image

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_decoder_run(struct ctr1474_ch* e, ctr1474_len_cb cbl, ctr1474_frame_in_cb cbf, int ad_hdr);


/*! \brief frees the resources asociated with an ctr1474 instance

  \param e the ctr1474 instance
*/
void simple_ctr1474_free(struct ctr1474_ch* e);

/*! \brief forces the stop of currently running encoding/decoding process

  Normally the encoding/decoding process should be stopped by using
  the return value of the user callback function. This function should
  be used only when that option is not possibile.

  \param e the ctr1474 instance
*/
void simple_ctr1474_force_stop(struct ctr1474_ch* e);


/*! \brief encodes/decodes a bunch of images using HIPI

  This function is not meant for interactive encoding/decoding of a
  video stream but for batch use like in a jpeg2000 hardware
  acceleration application.

  \param enc the ctr1474 instance for the encoder

  \param dec the ctr1474 instance for the decoder

  \param fw the firmware to load to the ADV212

  \param fwdata firmware description

  \param fwparams parameters for the firmware

  \param n the number of images to encode/decode

  \param in the array of pointers to source image bits

  \param ins the array with length of source images

  \param out the array of pointers to image bits created by the
  routine. The memory for these is allocated by this routine and
  should be deallocated by the user

  \param outs the array with length of images created by the routine

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_hipi(struct ctr1474_ch *enc, struct ctr1474_ch *dec,
		      char *fw, ETHJPG2K_FIRMWARE_DATA *fwdata, ETHJPG2K_FIRMWARE_PARAMS *fwparams,
		      int n,
		      char **in, int *ins,
		      char **out, int *outs);


/*! \brief returns the status of the input GPIO

  \param e the ctr1474 instance

  \return the value of the input GPIO bits
*/
int simple_ctr1474_gpio_read(struct ctr1474_ch* e);

/*! \brief returns the status of the input GPIO

  \param e the ctr1474 instance

  \param enable non-zero enables the play-back path

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_pb(struct ctr1474_ch* e, int enable);

/*! \brief writes to the output GPIO

  \param e the ctr1474 instance

  \param data the bits to be set on the output GPIO lines

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_gpio_write(struct ctr1474_ch* e, int data);

/*! \brief set the status of the 2 LEDS on the board

  \param e the ctr1474 instance

  \param data the bits or the 2 LEDS present on the CTR1474 board

  \return 0 on success, otherwise non-zero
*/
int simple_ctr1474_led(struct ctr1474_ch* e, int data);

/*! \brief reads the value of a ADV212 register

  \param e pointer to the ADV212 instance

  \param addr address of register to read

  \return value read from register
*/
int adv212_read_reg(struct ctr1474_ch *e, int addr);

/*! \brief writes to a ADV212 register

  \param e pointer to the ADV212 instance

  \param addr address of register to write

  \param data value to write
*/
void adv212_write_reg(struct ctr1474_ch *e, int addr, int data);

#endif
