
/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.

  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "version.h"
#include "libctr1474.h"

#define MAX_PATH 255

int is_ntsc = 0;

void usage(void) {
  printf("usage: test_decode [dev] [cdev] [n sec] [jp2 images .....]\n");
  exit(1);
}

/* path for the firmware */
char firmware[MAX_PATH]="firmware/decode_2_14_1.sea";

ETHJPG2K_FIRMWARE_DATA fwdata = {};

/* JPEG2000 decoder parameters. See ADV212 data sheet for more
   information */
ETHJPG2K_FIRMWARE_PARAMS fwparams = {
  /* these are values suggested by Christine Bako*/
  .DecodeParams= {
    .VFORMAT = 0,		/* will be automatically filled */
    .UNI = 3,
    .COD_STYLE = 2,		/* 1 is JC2, 2 is JP2 */
    .PICFG = 0x11,
  },
};

/* array of images */
struct image_s {
  char *data;
  int size;
} *images;

/* for every image keep track if this is an odd or even field */
int *oddeven;

/* needed to calculate the actual FPS */
static int cur, n, c=0;
static time_t first = 0;
static time_t finish_epoch;

/* get the content of the image */
void get_image(unsigned char *data, unsigned long len, int *field) {
#if 0
  struct timeval tv;

  gettimeofday(&tv, NULL);
  fprintf(stderr, "%d %ld.%ld\n", c, tv.tv_sec, tv.tv_usec);
#endif
  *field = oddeven[cur];
  memcpy(data, images[cur].data, len);
  cur = (cur + 1) % n;
  c++;
}

/* get the image len */
int get_len (void) {
  time_t last;

  if (time(&last) > finish_epoch) {
    if (c>0)
      printf("%f fps\n", ((float) c) / (last-first) );
    return -1;
  }
  if (!first)
    first = last;
  return images[cur].size;
}


int main(int argc, char *argv[]) {
  int i, j;
  char *dev;
  int coe = 0;;
  struct ctr1474_ch *e;

  printf("Eurotech CTR1474 utility, version " VERSION " compiled on " __DATE__ " " __TIME__ "\n");

  if (argc < 4) {
    usage();
  }
  dev = argv[1];

  if (is_ntsc)
    fwparams.DecodeParams.VFORMAT = 0;
  else
    fwparams.DecodeParams.VFORMAT = 1;

  n = argc - 4;
  images = calloc(n, sizeof(struct image_s));
  assert(images);
  oddeven = calloc(n, sizeof(int));
  assert(oddeven);
  j = 0;
  for(i=0; i<n; i++) {
    char *fname = argv[i+4];
    char *suffix;
    FILE *f;
    int padded;

    /* load the images to memory. If the stream is interlaced be
       careful to alternate an even and an old field. */
    suffix = strstr(fname, ".jp2");
    if (suffix) {
      if (suffix[-2] == '_') {
	int cur;

	cur = strtol(&suffix[-1], NULL, 0);
	if (coe == cur) {
	  coe = (coe + 1) & 1;
	  oddeven[j] = cur;
	}
	else
	  continue;		/* the for loop */
      }
      else {
	oddeven[j] = 0;
	if (is_ntsc)
	  fwparams.DecodeParams.VFORMAT = 8;
	else
	  fwparams.DecodeParams.VFORMAT = 9;
      }
    }
    else {
      fprintf(stderr, "Not a .jp2 file!\n");
      exit(1);
    }

    /* get the size and then load the image bits */
    f = fopen(fname, "r");
    if (!f) {
      fprintf(stderr, "Cannot open %s!\n", fname);
      exit(1);
    }
    fseek(f, 0, SEEK_END);
    images[j].size = ftell(f);
    assert(images[j].size);
    printf("image %d is %s, size %d\n", j, fname, images[j].size);
    rewind(f);
    padded = images[j].size;
    images[j].data = calloc(1, padded);
    assert(images[j].data);
    fread(images[j].data, images[j].size, 1, f);
    fclose(f);
    images[j].size = padded;
    j++;
  }
  n = j;
  assert(n);

  /* initialize the decoder */
  e = simple_ctr1474(argv[1], argv[2]);
  assert(e);
  simple_ctr1474_pb(e, 1);
  assert(!simple_ctr1474_board_init(e, is_ntsc, firmware, &fwdata, &fwparams));

  /* put the video output in one channel playback */
  twa2835_mon_full(TWA(e), 0);
  twa2835_mon_pb_enable(TWA(e), 0, 1);

  finish_epoch = time(NULL) + strtoul(argv[3], NULL, 0);
  cur = 0;
  simple_ctr1474_decoder_run(e, get_len, get_image, 1);

  /* restore monitoring mode */
  twa2835_mon_pb_enable(TWA(e), 0, 0);

  return 0;
}
